/************************************************************
 *    Copyright (C) 2008-2009                               *
 *    Michel Ludwig (michel.ludwig@liverpool.ac.uk)         *
 *    University of Liverpool                               *
 *                                                          *
 *    This program is free software; you can redistribute   *
 *    it and/or modify it under the terms of the GNU        *
 *    General Public License as published by the Free       *
 *    Software Foundation; either version 3 of the License, *
 *    or (at your option) any later version.                *
 *                                                          *
 *    This program is distributed in the hope that it will  *
 *    be useful, but WITHOUT ANY WARRANTY; without even     *
 *    the implied warranty of MERCHANTABILITY or FITNESS    *
 *    FOR A PARTICULAR PURPOSE.  See the GNU General Public *
 *    License for more details.                             *
 *                                                          *
 *    You should have received a copy of the GNU General    *
 *    Public License along with this program; if not, see   *
 *    <http://www.gnu.org/licenses/>.                       *
 ************************************************************/
%{

#include <stdlib.h>
#include <stdio.h>
#include <string.h>  
#include "tree.h"
#include "symbol_table.h"
  
/* flex functions */

extern int yylex () ;

/* not using the flex function, yyerror was redefined */

void yyerror (char const *s);
 
/* Max stack size for parsing */
  
#define YYMAXDEPTH 136854775807
    
 extern list *f_list;
 extern tree *root;
 
// forward declaration
 
 extern int numtokens;
 extern int numlines;
 extern int numcolumns;

 /* global variables for parser.y */

 int numerrors = 0;  

%}

%define parse.error verbose
%debug
%defines

%token FALSE TRUE NAME TNOT TAND TOR TIMPLY TIFF TONLYIF TNEXT TALWAYS TSOMETIME TUNTIL TUNLESS LEFT_BRACKET RIGHT_BRACKET TCOMMA

%union {
  char *strValue;
  struct tree* ctree;
  struct list* clist;
}

%type <ctree> input formula unary_formula binary_formula n_ary_formula
%type <strValue> NAME TRUE FALSE TAND TOR TNOT TALWAYS TNEXT TSOMETIME LEFT_BRACKET RIGHT_BRACKET TUNLESS TUNTIL

%right TUNLESS TUNTIL TIMPLY TIFF TONLYIF
%right TOR TAND 
%right TALWAYS TNEXT TSOMETIME
%right TNOT
%right LEFT_BRACKET RIGHT_BRACKET

%%

input: formula
       { root = $1;
	 root->parent = NULL;
       };

formula:
	LEFT_BRACKET formula RIGHT_BRACKET
	{
	  $$ = $2;
	}
	| unary_formula
	{
	  $$ = $1;
	}
	| binary_formula
	{
	  $$ = $1;
	}
        | n_ary_formula
	{
	  $$ = $1;
	}
        | TRUE
	{
          tree *t = new_tree();
	  t = tree_op(t,CONSTANT);
	  t->index = st_InsertEntry("true");
	  t->index = CTRUE;
	  //	  free($1);
	  $$ = t;
	}
        | FALSE
	{
          tree *t = new_tree();
	  t = tree_op(t,CONSTANT);
	  t->index = st_InsertEntry("false");
	  t->index = CFALSE;
	  //free($1);
	  $$ = t;
	}
        | NAME
	{
	  tree *t = new_tree();
	  t = tree_op(t,PROP);
	  t->index = st_InsertEntry($1);
	  free($1);
	  $$ = t;
	}
       ;


binary_formula:
	formula TUNTIL formula
	{
	  tree *t = new_tree();
	  t = tree_op(t,UNTIL);
	  t = tree_children(t,$1,$3,NULL);
	  $$ = t;
	}
        | formula TUNLESS formula
	{
	  tree *t = new_tree();
	  t = tree_op(t,UNLESS);
	  t = tree_children(t,$1,$3,NULL);
	  $$ = t;
	}
	| formula TIMPLY formula
	{
	  tree *t = new_tree();
	  t = tree_op(t,IMPLY);
	  t = tree_children(t,$1,$3,NULL);
	  $$ = t;
	}
	| formula TONLYIF formula
	{
	  tree *t = new_tree();
	  t = tree_op(t,IMPLY);
	  t = tree_children(t,$3,$1,NULL);
	  $$ = t;
	}
	| formula TIFF formula
	{
	  tree *t = new_tree();
	  t = tree_op(t,IFF);
	  t = tree_children(t,$1,$3,NULL);
	  $$ = t;
	};
	
unary_formula:
        TNOT formula
	{
	  tree *t = new_tree();
	  t = tree_op(t,NOT);
	  t = tree_children(t,$2,NULL,NULL);
	  $$ = t;
	}
	| TALWAYS formula
	{
	  tree *t = new_tree();
	  t = tree_op(t,ALWAYS);
	  t = tree_children(t,$2,NULL,NULL);
	  $$ = t;
	}
	| TNEXT formula
	{
	  tree *t = new_tree();
	  t = tree_op(t,NEXT);
	  t = tree_children(t,$2,NULL,NULL);
	  $$ = t;
	}
	| TSOMETIME formula
	{
	  tree *t = new_tree();
	  t = tree_op(t,SOMETIME);
	  t = tree_children(t,$2,NULL,NULL);
	  $$ = t;
	};
	

n_ary_formula:
              formula TAND formula
	      {
		tree *t = new_tree();
		t = tree_op(t,AND);
		list *l = tree_to_list($1,$3,AND);
		t = tree_children(t,NULL,NULL,l);
		$$ = t;
	      }
              | formula TOR formula
	      {
		tree *t = new_tree();
		t = tree_op(t,OR);
		list *l = tree_to_list($1,$3,OR);
		t = tree_children(t,NULL,NULL,l);
		$$ = t;
	      };	      

%%

list *f_list = NULL;

void yyerror (char const *s) {
  fprintf (stderr, "\n Error: %s, line %d, column %d\n",s,numlines,numcolumns);
  numerrors++;
}
