/************************************************************
 *    Copyright (C) 2008-2009                               *
 *    Michel Ludwig (michel.ludwig@liverpool.ac.uk)         *
 *    University of Liverpool                               *
 *                                                          *
 *    This program is free software; you can redistribute   *
 *    it and/or modify it under the terms of the GNU        *
 *    General Public License as published by the Free       *
 *    Software Foundation; either version 3 of the License, *
 *    or (at your option) any later version.                *
 *                                                          *
 *    This program is distributed in the hope that it will  *
 *    be useful, but WITHOUT ANY WARRANTY; without even     *
 *    the implied warranty of MERCHANTABILITY or FITNESS    *
 *    FOR A PARTICULAR PURPOSE.  See the GNU General Public *
 *    License for more details.                             *
 *                                                          *
 *    You should have received a copy of the GNU General    *
 *    Public License along with this program; if not, see   *
 *    <http://www.gnu.org/licenses/>.                       *
 ************************************************************/
 
%{

#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include "parser.tab.h"

/* Global Variables */

int numtokens = 0;
int numlines = 1;
int numcolumns = 1;

%}


%option noyywrap

AND           "^"|"&"|(?i:and)
OR            "+"|"|"|(?i:or)|(?i:v)|\x7c
NOT           "!"|"-"|"~"|(?i:not)|(?i:neg)
IFF           "<->"|"<=>"|(?i:iff)
ONLYIF        "<-"|"<="|(?i:onlyif)
IMPLY         "->"|"=>"|(?i:imp)|(?i:imply)|(?i:implies)
ALWAYS        "[]"|(?i:always)
NEXT          "O"|(?i:next)
SOMETIME      "<>"|(?i:sometime)
UNTIL         "U"|(?i:until)
UNLESS        "W"|(?i:unless)
TRUE          (?i:true)
FALSE         (?i:false)
NAME          [[:alpha:]][[:alnum:]_]*
LEFT_BRACKET  "("|"["
RIGHT_BRACKET ")"|"]"
COMMA         ","
WS            [ \t\v\f\$]
EOL           [\n\r]
%x            BLOCK_COMMENT
%x            LINE_COMMENT

%%

{AND}        { numtokens++; numcolumns+=yyleng; return TAND; }
{OR}         { numtokens++; numcolumns+=yyleng; return TOR; }
{NOT}        { numtokens++; numcolumns+=yyleng; return TNOT; }
{IMPLY}      { numtokens++; numcolumns+=yyleng; return TIMPLY; }
{ONLYIF}     { numtokens++; numcolumns+=yyleng; return TONLYIF; }
{IFF}        { numtokens++; numcolumns+=yyleng; return TIFF; }
{ALWAYS}     { numtokens++; numcolumns+=yyleng; return TALWAYS; }
{NEXT}       { numtokens++; numcolumns+=yyleng; return TNEXT; }
{SOMETIME}   { numtokens++; numcolumns+=yyleng; return TSOMETIME; }
{UNTIL}      { numtokens++; numcolumns+=yyleng; return TUNTIL; }
{UNLESS}     { numtokens++; numcolumns+=yyleng; return TUNLESS; }

{LEFT_BRACKET}  { numtokens++; numcolumns+=yyleng; return LEFT_BRACKET; }
{RIGHT_BRACKET} { numtokens++; numcolumns+=yyleng; return RIGHT_BRACKET; }

{TRUE}       { numtokens++; numcolumns+=yyleng; return TRUE; }
{FALSE}       { numtokens++; numcolumns+=yyleng; return FALSE; }

{NAME}       { numtokens++; numcolumns+=yyleng;  
               yylval.strValue = strdup(yytext); 
               int i = 0; 
               for (i=0;i<=yyleng;i++) yylval.strValue[i]=tolower(yylval.strValue[i]); 
               return NAME; 
             }
{COMMA}      { numtokens++; numcolumns++; return TCOMMA; }
"."          { numtokens++; numcolumns++; }

{EOL}        { numlines++; numcolumns=1; }
{WS}         { numcolumns++; }

"/*"                 { BEGIN (BLOCK_COMMENT); }
<BLOCK_COMMENT>"*/"  { BEGIN (INITIAL); }
<BLOCK_COMMENT>{EOL} { ++ numlines; numcolumns = 1; }
<BLOCK_COMMENT>.     { ++ numcolumns; }

"//"|"%"            { BEGIN (LINE_COMMENT); }
<LINE_COMMENT>{EOL} { ++ numlines; numcolumns = 1; BEGIN (INITIAL); }
<LINE_COMMENT>.     { ++ numcolumns; }
%%
